// Simple collision detection example
// Moving spheres in polyhedra world, using planes only (no detection of edges and corners)
// which works pretty well. Static models are given in world coordinates to avoid transformations.

// Interactive version of the most basic collision detection, no ball-to-ball collision, no gravity,
// no speed adjustment in collisions. You just move the ball and if it intersects with the octotoad,
// it is moved out from it.

// gcc collision0-interactive.c -o collision0-interactiv ../common/*.c ../common/Linux/*.c -I../common -I../common/Linux -lGL -lX11 -DGL_GLEXT_PROTOTYPES -lm

#define MAIN
#include "GL_utilities.h"
#include "VectorUtils4.h"
#include "LittleOBJLoader.h"
#include "MicroGlut.h"
//uses framework Cocoa
//uses framework OpenGL

mat4 projectionMatrix;

Model *octotoad;
Model *sphere;

vec3 objectcolor = vec3(1.0, 1.0, 0.5);
vec3 ballcolor = vec3(1.0, 0.4, 0.2);

// Array of balls (world coords)
#define kNumBalls 1
vec3 pos[kNumBalls];
vec3 speed[kNumBalls];

// Reference to shader program
GLuint program;

void init(void)
{
	// GL inits
	glClearColor(0.4,0.4,0.7,0);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	printError("GL inits");

	// Load and compile shader
	program = loadShaders("phong.vert", "phong.frag");
	glUseProgram(program);
	printError("init shader");
	
	// Upload geometry to the GPU:
	octotoad = LoadModel("data/octotoad.obj");
	sphere = LoadModel("data/sphere20.obj");

	projectionMatrix = frustum(-0.6, 0.6, -0.6, 0.6, 1, 30);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);
	
	pos[0] = vec3(0, 2, 0);
	speed[0] = vec3(0, 0, 0);
	
	printError("init arrays");
}

GLfloat a = 0.0;

void SimpleCollisionTest(vec3 *pos, vec3 *speed, GLfloat radius, Model *m)
{
	int i;
	vec3 p0, p1, p2, v1, v2, n;
	GLfloat diff;
	
	GLfloat bestDiff = -10000000; // Mycket
	vec3 bestMove;
	
// For all polygons in model
// Is there an overlap? Save the smallest!
	for (i = 0; i < m->numIndices-1; i+=3) // 3 indices = triangle
	{
		// Calc normal vector
		p0 = m->vertexArray[m->indexArray[i]]; // Indexing was changed (for the better) compared to loadobj.
		p1 = m->vertexArray[m->indexArray[i+1]];
		p2 = m->vertexArray[m->indexArray[i+2]];
		v1 = p1 - p0;
		v2 = p2 - p0;
		n = cross(v1, v2); // Normal vector pointing out
		n = Normalize(n);
		// Dot products with pos and vertex in triangle
		// Compare with radius
		diff = - dot(n, p0) - radius + dot(n, *pos);
		if (diff > bestDiff)
		{
			bestDiff = diff;
			bestMove = -diff*n;
		}
	}
	if (bestDiff < 0)
	{
		printf("Hit with diff %f offsets %1.2f %1.2f %1.2f\n", bestDiff, bestMove.x, bestMove.y, bestMove.z);
		*pos = bestMove + *pos;
	}
	else
		printf("Miss with diff %f\n", bestDiff);		
}

void SpheresToSpheresCollision()
{
	// See other demos
}

void display(void)
{
	int i;
	Matrix4D modeltoworld, worldtoview, m;
	
	printError("pre display");

	// clear the screen
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	worldtoview = lookAt(-1, 3, -3, 0,0,0, 0,1,0);
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, worldtoview.m);
	
	glUseProgram(program);
	glUniform3fv(glGetUniformLocation(program, "color"), 1, (GLfloat *)&objectcolor);
	DrawModel(octotoad, program, "inPosition", "inNormal", "");

	glUniform3fv(glGetUniformLocation(program, "color"), 1, (GLfloat *)&ballcolor);
	for (i = 0; i < kNumBalls; i++)
	{
		SimpleCollisionTest(&pos[i], &speed[i], 0.2, octotoad);
		
		modeltoworld = T(pos[i].x, pos[i].y, pos[i].z);
		m = worldtoview * modeltoworld;
	
		glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
		glUseProgram(program);
		DrawModel(sphere, program, "inPosition", "inNormal", "");
	}
	
	printError("display");
	
	glutSwapBuffers();
}

void keys(unsigned char key, int x, int y)
{
	switch (key)
	{
		case 'w': pos[0].y += 0.1;break;
		case 'a': pos[0].x += 0.1;break;
		case 's': pos[0].y -= 0.1;break;
		case 'd': pos[0].x -= 0.1;break;
	}
	glutPostRedisplay();
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitContextVersion(3,2);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow ("interactive collision - move ball with wasd");
	glutDisplayFunc(display); 
	glutKeyboardFunc(keys);
	init ();
	glutMainLoop();
	exit(0);
}
