// Simple collision detection example
// Moving spheres in polyhedra world, using planes only (no detection of edges and corners)
// which works pretty well. Static models are given in world coordinates to avoid transformations.
// Case 2, three sloping blocks. No invible "box" but respawns balls instead. Otherwise the same.

// LOTS of improvements are possible and desirable. For example, multiple collisions in the same
// frame should be handled a lot better. You should not change the position and speed, but rather
// accumulate changes and apply a total.

// Polyhedra-polyhedra tests are more complex. The simplest is probably the volume intersection method,
// although that isn't the fastest. Faster methods need additional data to give connectivity, like finding
// all polygons used by a certain vertex. (Great for "walking over the surface" looking for the right
// side to start.)

// But we are not trying to make a full-fledged physics engine here, but a small, digestable intro
// to the basics of collision detection (running in an OpenGL 3 environment).

// By Ingemar Ragnemalm 2013
// C++ version 2024

// SAME AS collision2-surfaces except that the two surfaces now share one OBJ file! This means that you can build a scene from multiple parts as one model! (You only need a modeller that can export a model that way.)

// gcc collision2-surfaces-multiobj.c -o collision2-surfaces-multiobj ../common/*.c ../common/Linux/*.c -I../common -I../common/Linux -lGL -lX11 -DGL_GLEXT_PROTOTYPES -lm

#define MAIN
#include "GL_utilities.h"
#include "VectorUtils4.h"
#include "LittleOBJLoader.h"
#include "MicroGlut.h"
//uses framework Cocoa
//uses framework OpenGL

mat4 projectionMatrix;

Model *sphere; // Balls
Model **things;
vec3 mojangcolor = vec3(0.5, 0.7, 0.4);

// Array of balls (world coords)
#define kNumBalls 10
vec3 pos[kNumBalls];
vec3 speed[kNumBalls];
vec3 color[kNumBalls];
vec3 grav = vec3(0, -0.01, 0);

#define kBounce 0.7

// Reference to shader program
GLuint program;

void init(void)
{
	// GL inits
	glClearColor(0.2,0.2,0.5,0);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	printError("GL inits");

	// Load and compile shader
	program = loadShaders("phong.vert", "phong.frag");
	glUseProgram(program);
	printError("init shader");
	
	// Upload geometry to the GPU:

	things = LoadModelSet("data/surfaces.obj");
	
	sphere = LoadModelPlus("data/sphere20.obj");

	projectionMatrix = frustum(-0.6, 0.6, -0.6, 0.6, 1, 30);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);

	pos[0] = SetVector(0, 2, 0);
	speed[0] = SetVector(0, -0.2, 0);
	color[0] = SetVector(1, 1, 1); // white
	
	pos[1] = SetVector(-1, 0, 0);
	speed[1] = SetVector(0.5, 0, 0);
	color[1] = SetVector(1, 0, 0); // red

	pos[2] = SetVector(0, 1, 0);
	speed[2] = SetVector(0, 1, 0);
	color[2] = SetVector(0, 1, 0); // green

	pos[3] = SetVector(1, 1, 1);
	speed[3] = SetVector(0, 1, 0);
	color[3] = SetVector(0, 0, 1); // blue

	pos[4] = SetVector(0, 1, 0.5);
	speed[4] = SetVector(0.2, 0.3, 0.4);
	color[4] = SetVector(1, 1, 0); // yellow

	pos[5] = SetVector(1, 1, 0.5);
	speed[5] = SetVector(0, 0, 0);
	color[5] = SetVector(1, 0, 1); // Magenta

	pos[6] = SetVector(-1, 1, 0);
	speed[6] = SetVector(0, 0, 0);
	color[6] = SetVector(0, 1, 1); // Cyan

	pos[7] = SetVector(0.5, 0, 0.5);
	speed[7] = SetVector(0, 0, 0);
	color[7] = SetVector(0.5, 0.5, 0.5); // Gray

	pos[8] = SetVector(-0.5, 0, 0.5);
	speed[8] = SetVector(0, 0, 0);
	color[8] = SetVector(1, 0.5, 0); // Orange

	pos[9] = SetVector(0.5, 1, 0);
	speed[9] = SetVector(0, 0, 0);
	color[9] = SetVector(0.5, 1, 0); // Light green
}

GLfloat a = 0.0;

void SimpleCollisionTest(vec3 *pos, vec3 *speed, GLfloat radius, Model *m)
{
	int i;
	vec3 p0, p1, p2, v1, v2, n;
	GLfloat diff;
	
	GLfloat bestDiff = -10000000; // A lot
	vec3 bestMove, bestNormal;
	
// For all polygons in model
// Is there an overlap? Save the smallest!
	for (i = 0; i < m->numIndices-1; i+=3) // 3 indices = triangle
	{
		// Calc normal vector
		p0 = m->vertexArray[m->indexArray[i]]; // Indexing was changed (for the better) compared to loadobj.
		p1 = m->vertexArray[m->indexArray[i+1]];
		p2 = m->vertexArray[m->indexArray[i+2]];
		v1 = VectorSub(p1, p0);
		v2 = VectorSub(p2, p0);
		n = CrossProduct(v1, v2); // Normal vector pointing out
		n = Normalize(n);
		// Dot products with pos and vertex in triangle
		// Compare with radius
		diff = - DotProduct(n, p0) - radius + DotProduct(n, *pos);
		if (diff > bestDiff)
		{
			bestDiff = diff;
			bestMove = ScalarMult(n, -diff);
			bestNormal = n;
		}
	}
	if (bestDiff < 0)
	if (bestDiff > -10000000)
	{
		*pos = VectorAdd(bestMove, *pos);
		
		// Change speed!
		float ns = DotProduct(*speed, bestNormal);
		bestNormal = ScalarMult(bestNormal, -(1+kBounce)*ns); // Change of speed
		*speed = VectorAdd(bestNormal, *speed);
	}
}

// Simplified solution: No variation in weight, no friction, no rotation...
// We do this "for real" in TSBK03 with a physically-based model.
void SpheresToSpheresCollision(float radius)
{
	int i, j;
	
	for (i = 0; i < kNumBalls; i++)
		for (j = i+1; j < kNumBalls; j++)
		{
			vec3 diff, diffV;
			
			diff = VectorSub(pos[i], pos[j]); // Position difference
			diffV = VectorSub(speed[i], speed[j]); // Speed difference

			if (Norm(diff) < 2*radius) // Close enough to collide?
			{
				// Separate!
				float len = Norm(diff) - 2*radius; // This is negative!
				vec3 separate = diff;
				separate = Normalize(separate);
				separate = ScalarMult(separate, len * 0.5);
				pos[i] = VectorSub(pos[i], separate);
				pos[j] = VectorAdd(pos[j], separate);
				
				// Change speeds
				if (DotProduct(diff, diffV) < 0.0) // Moving towards each other? Otherwise we don't change speed!
				{
					diff = Normalize(diff);
					float ns = DotProduct(diff, diffV);
					diff = ScalarMult(diff, ns * (0.5 + kBounce/2));
					
					speed[i] = VectorSub(speed[i], diff); // Add same change to both, simplified impulse
					speed[j] = VectorAdd(speed[j], diff);
				}
			}
		}
}

void display(void)
{
	int i, j;
	Matrix4D modeltoworld, worldtoview, m;
	
	printError("pre display");

	// clear the screen
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	worldtoview = lookAt(-1,3,-4, -1.2,0,0, 0,1,0);
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, worldtoview.m);
	
	glUseProgram(program);
	glUniform3fv(glGetUniformLocation(program, "color"), 1, (GLfloat *)&mojangcolor);
	for (i = 0; things[i] != NULL; i++)
		DrawModel(things[i], program, "inPosition", "inNormal", "");

	SpheresToSpheresCollision(0.2);
	
	for (i = 0; i < kNumBalls; i++)
	{
		speed[i] = VectorAdd(grav, speed[i]);
		pos[i] = VectorAdd(pos[i], speed[i]);
		
		// If the balls leave the world, respawn
		if (pos[i].y < -10)
		{
			pos[i].y = 3 + (float)rand()/RAND_MAX;
			pos[i].x = -2 + (float)rand()/RAND_MAX-0.5;
			pos[i].z = 0 + (float)rand()/RAND_MAX-0.5;
			speed[i] = SetVector(0, 0, 0);
		}

		for (j = 0; things[j] != NULL; j++)
			SimpleCollisionTest(&pos[i], &speed[i], 0.2, things[j]);
		
		modeltoworld = T(pos[i].x, pos[i].y, pos[i].z);
		m = Mult(worldtoview, modeltoworld); // tr * ry * rz
	
		glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
		glUseProgram(program);
		glUniform3fv(glGetUniformLocation(program, "color"), 1, (GLfloat *)&color[i]);
		DrawModel(sphere, program, "inPosition", "inNormal", "");
	}
	
	printError("display");
	
	glutSwapBuffers();
}

void keys(unsigned char key, int x, int y)
{
	switch (key)
	{
		case 'w': pos[0].y += 0.1;break;
		case 'a': pos[0].x += 0.1;break;
		case 's': pos[0].y -= 0.1;break;
		case 'd': pos[0].x -= 0.1;break;
	}
	glutPostRedisplay();
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitContextVersion(3,2);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(500,500);
	glutCreateWindow ("simple collision example 2 (multi-obj)");
	glutRepeatingTimer(20);
	glutDisplayFunc(display); 
	glutKeyboardFunc(keys);
	init ();
	glutMainLoop();
	exit(0);
}
