// Simple collision detection example
// Moving spheres in polyhedra world, using planes only (no detection of edges and corners)
// which works pretty well. Static models are given in world coordinates to avoid transformations.

// Variant 4: Nasty. I tried to construct a typical hard case, one that is challenging even for
// many serious physics enginges... I expected my primitive example to run into problems, but no.
// The result is almost embarassingly good. Two different pairs of planes are included, one that
// works quite well and one where we can see some problems. (Select by changing a constant below.)

// LOTS of improvements are possible and desirable. For example, multiple collisions in the same
// frame should be handled a lot better. You should not change the position and speed, but rather
// accumulate changes and apply a total.

// Polyhedra-polyhedra tests are more complex. The simplest is probably the volume intersection method,
// although that isn't the fastest. Faster methods need additional data to give connectivity, like finding
// all polygons used by a certain vertex. (Great for "walking over the surface" looking for the right
// side to start.)

// But we are not trying to make a full-fledged physics engine here, but a small, digestable intro
// to the basics of collision detection (running in an OpenGL 3 environment).

// By Ingemar Ragnemalm 2013
// C++ version 2024

// gcc collision4-nasty.c -o collision4-nasty ../common/*.c ../common/Linux/*.c -I../common -I../common/Linux -lGL -lX11 -DGL_GLEXT_PROTOTYPES -lm

#include <stdlib.h>
#define MAIN
#include "GL_utilities.h"
#include "VectorUtils4.h"
#include "LittleOBJLoader.h"
#include "MicroGlut.h"
//uses framework Cocoa
//uses framework OpenGL

mat4 projectionMatrix;

Model *nasty[2]; // Static thing to bounce off.
Model *sphere; // Balls
vec3 objectcolor[2] = {vec3(0.7, 0.9, 1.0),
						vec3(0.9, 0.7, 1.0)};

// Array of balls (world coords)
#define kNumBalls 10
vec3 pos[kNumBalls];
vec3 speed[kNumBalls];
vec3 color[kNumBalls];
vec3 grav = vec3(0, -0.01, 0);

#define kBounce 0.7
// Set this to 0 or 1:
#define harderCase 0

// Reference to shader program
GLuint program;

void init(void)
{
	// GL inits
	glClearColor(0.2,0.2,0.5,0);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	printError("GL inits");
	
	// Load and compile shader
	program = loadShaders("phong.vert", "phong.frag");
	glUseProgram(program);
	printError("init shader");
	
	// Upload geometry to the GPU:
	if (!harderCase)
	{
		// Two less steep planes - works surprisingly well
		nasty[0] = LoadModel("data/nastier1.obj");
		nasty[1] = LoadModel("data/nastier2.obj");
	}
	else
	{
		// Two steeper planes - visible artifacts! (But I have seen worse.)
		nasty[0] = LoadModel("data/nastierb1.obj");
		nasty[1] = LoadModel("data/nastierb2.obj");
	}
	sphere = LoadModelPlus("data/sphere20.obj");

	projectionMatrix = frustum(-0.6, 0.6, -0.6, 0.6, 1, 30);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);

	pos[0] = vec3(0, 2, 0);
	speed[0] = vec3(0, -0.2, 0);
	color[0] = vec3(1, 1, 1); // white
	
	pos[1] = vec3(-1, 0, 0);
	speed[1] = vec3(0.5, 0, 0);
	color[1] = vec3(1, 0, 0); // red

	pos[2] = vec3(0, 1, 0);
	speed[2] = vec3(0, 1, 0);
	color[2] = vec3(0, 1, 0); // green

	pos[3] = vec3(1, 1, 1);
	speed[3] = vec3(0, 1, 0);
	color[3] = vec3(0, 0, 1); // blue

	pos[4] = vec3(0, 1, 0.5);
	speed[4] = vec3(0.2, 0.3, 0.4);
	color[4] = vec3(1, 1, 0); // yellow

	pos[5] = vec3(1, 1, 0.5);
	speed[5] = vec3(0, 0, 0);
	color[5] = vec3(1, 0, 1); // Magenta

	pos[6] = vec3(-1, 1, 0);
	speed[6] = vec3(0, 0, 0);
	color[6] = vec3(0, 1, 1); // Cyan

	pos[7] = vec3(0.5, 0, 0.5);
	speed[7] = vec3(0, 0, 0);
	color[7] = vec3(0.5, 0.5, 0.5); // Gray

	pos[8] = vec3(-0.5, 0, 0.5);
	speed[8] = vec3(0, 0, 0);
	color[8] = vec3(1, 0.5, 0); // Orange

	pos[9] = vec3(0.5, 1, 0);
	speed[9] = vec3(0, 0, 0);
	color[9] = vec3(0.5, 1, 0); // Light green

	printError("init arrays");
}

GLfloat a = 0.0;

void SimpleCollisionTest(vec3 *pos, vec3 *speed, GLfloat radius, Model *m)
{
	int i;
	vec3 p0, p1, p2, v1, v2, n;
	GLfloat diff;
	
	GLfloat bestDiff = -10000000; // A lot
	vec3 bestMove, bestNormal;
	
// For all polygons in model
// Is there an overlap? Save the smallest!
	for (i = 0; i < m->numIndices-1; i+=3) // 3 indices = triangle
	{
		// Calc normal vector
		p0 = m->vertexArray[m->indexArray[i]]; // Indexing was changed (for the better) compared to loadobj.
		p1 = m->vertexArray[m->indexArray[i+1]];
		p2 = m->vertexArray[m->indexArray[i+2]];
		v1 = p1 - p0;
		v2 = p2 - p0;
		n = cross(v1, v2); // Normal vector pointing out
		n = Normalize(n);
		// Dot products with pos and vertex in triangle
		// Compare with radius
		diff = - dot(n, p0) - radius + dot(n, *pos);
		if (diff > bestDiff)
		{
			bestDiff = diff;
			bestMove = -diff * n;
			bestNormal = n;
		}
	}
	if (bestDiff < 0)
	if (bestDiff > -10000000)
	{
		*pos = bestMove + *pos;
		
		// Change speed!
		float ns = dot(*speed, bestNormal);
		bestNormal = bestNormal * -(1+kBounce)*ns; // Change of speed
		*speed = bestNormal + *speed;
	}
}

void SpheresToSpheresCollision(float radius)
{
	int i, j;
	
	for (i = 0; i < kNumBalls; i++)
		for (j = i+1; j < kNumBalls; j++)
		{
			vec3 diff, diffV;
			
			diff = pos[i] - pos[j]; // Position difference
			diffV = speed[i] - speed[j]; // Speed difference

			if (Norm(diff) < 2*radius) // Close enough to collide?
			{
				// Separate!
				float len = Norm(diff) - 2*radius; // This is negative!
				vec3 separate = diff;
				separate = Normalize(separate);
				separate = separate * len * 0.5;
				pos[i] = pos[i] - separate;
				pos[j] = pos[j] + separate;
				
				// Change speeds
				if (dot(diff, diffV) < 0.0) // Moving towards each other? Otherwise we don't change speed!
				{
					diff = Normalize(diff);
					float ns = dot(diff, diffV);
					diff = diff * ns * (0.5 + kBounce/2);
					
					speed[i] = (speed[i] - diff); // Add same change to both, simplified impulse
					speed[j] = (speed[j] + diff);
				}
			}
		}
}

void display(void)
{
	int i;
	Matrix4D modeltoworld, worldtoview, m;
	
	printError("pre display");

	// clear the screen
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	worldtoview = lookAt(-0.2,2,-4, 0,0,0, 0,1,0);
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, worldtoview.m);
	
	glUseProgram(program);
	for (i = 0; i < 2; i++)
	{
		glUniform3fv(glGetUniformLocation(program, "color"), 1, (GLfloat *)&objectcolor[i]);
		DrawModel(nasty[i], program, "inPosition", "inNormal", "");
	}
	
	SpheresToSpheresCollision(0.2);
	
	for (i = 0; i < kNumBalls; i++)
	{
		speed[i] = VectorAdd(grav, speed[i]);
		pos[i] = VectorAdd(pos[i], speed[i]);
		
		// If the balls leave the world, respawn
		if (pos[i].y < -10)
		{
			pos[i].y = 3 + (float)rand()/RAND_MAX;
			pos[i].x = -0 + (float)rand()/RAND_MAX-0.5;
			pos[i].z = 0 + (float)rand()/RAND_MAX-0.5;
			speed[i] = vec3(0, 0, 0);
		}
		
		SimpleCollisionTest(&pos[i], &speed[i], 0.2, nasty[0]);
		SimpleCollisionTest(&pos[i], &speed[i], 0.2, nasty[1]);
		
		modeltoworld = T(pos[i].x, pos[i].y, pos[i].z);
		m = Mult(worldtoview, modeltoworld); // tr * ry * rz
	
		glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
		glUseProgram(program);
		glUniform3fv(glGetUniformLocation(program, "color"), 1, (GLfloat *)&color[i]);
		DrawModel(sphere, program, "inPosition", "inNormal", "");
	}
	
	printError("display");
	
	glutSwapBuffers();
}

void keys(unsigned char key, int x, int y)
{
	switch (key)
	{
		case ' ': pos[0].y += 2;break;
		case 'w': pos[0].z += 0.1;break;
		case 'a': pos[0].x += 0.1;break;
		case 's': pos[0].z -= 0.1;break;
		case 'd': pos[0].x -= 0.1;break;
	}
	glutPostRedisplay();
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitContextVersion(3,2);
	glutInitWindowSize(500,500);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow ("simple collision example 4");
	glutRepeatingTimer(20);
	glutDisplayFunc(display); 
	glutKeyboardFunc(keys);
	init ();
	glutMainLoop();
	exit(0);
}
